using System;
using System.IO;
using System.ComponentModel;

namespace id.Quake.Files {
	public partial class Level {

		#region Properties

		private readonly Node[] nodes;
		/// <summary>Gets the level's BSP nodes.</summary>
		[Category("BSP Tree")]
		[Description("All of the nodes in the level.")]
		public Node[] Nodes {
			get { return this.nodes; }
		}

		#endregion

		/// <summary>Describes a BSP node.</summary>
		[TypeConverter(typeof(ExpandableObjectConverter))]
		public class Node {

			private readonly Level Owner;

			private readonly uint LeafTestMask;
			private readonly uint LeafBitMask;

			#region Properties

			private readonly uint plane;
			/// <summary>Index of the plane that splites the node.</summary>
			public uint Plane {
				get { return this.plane; }
			}

			private readonly uint front;
			/// <summary>Index of front child node - or, if the most significant bit is set, ~index of child leaf.</summary>
			public uint Front {
				get { return this.front; }
			}

			/// <summary>Returns true if the front of this node is another node.</summary>
			public bool FrontIsNode { get { return (this.Front & this.LeafTestMask) == 0; } }

			/// <summary>Gets the index of the node or leaf to the front of this node.</summary>
			public uint FrontIndex { get { return this.FrontIsNode ? this.Front : (this.Front ^ this.LeafBitMask); } }

			private uint back;
			/// <summary>Index of back child node - or, if the most significant bit is set, ~index of child leaf.</summary>
			public uint Back {
				get { return this.back; }
			}

			/// <summary>Returns true if the back of this node is another node.</summary>
			public bool BackIsNode { get { return (this.Back & this.LeafTestMask) == 0; } }

			/// <summary>Gets the index of the node or leaf to the back of this node.</summary>
			public uint BackIndex { get { return this.BackIsNode ? this.Back : (this.Back ^ this.LeafBitMask); } }

			private readonly BoundingBoxShort bounds;
			/// <summary>The bounding box of this node and all of its children.</summary>
			public BoundingBoxShort Bounds {
				get { return this.bounds; }
			}

			private readonly ushort faceIndex;
			/// <summary>Index of the first face.</summary>
			public ushort FaceIndex {
				get { return this.faceIndex; }
			}

			private ushort faceCount;
			/// <summary>How many faces make up this node.</summary>
			public ushort FaceCount {
				get { return this.faceCount; }
			}

			#endregion


			#region Constructor

			public Node(Level level, Stream stream) {
				this.Owner = level;

				BinaryReader WorkingReader = new BinaryReader(stream);

				this.LeafTestMask = this.Owner.FileVersion.GameVersion == Version.Game.Quake ? 0x8000 : 0x80000000;
				this.LeafBitMask = this.Owner.FileVersion.GameVersion == Version.Game.Quake ? 0xFFFF : 0xFFFFFFFF;


				this.plane = WorkingReader.ReadUInt32();

				switch (this.Owner.FileVersion.GameVersion) {
					case Version.Game.Quake:
						this.front = WorkingReader.ReadUInt16();
						this.back = WorkingReader.ReadUInt16();
						break;
					case Version.Game.Quake2:
						this.front = WorkingReader.ReadUInt32();
						this.back = WorkingReader.ReadUInt32();
						break;
					default:
						throw new InvalidOperationException();
				}

				this.bounds = new BoundingBoxShort(new Vector3Short(WorkingReader.ReadInt16(), WorkingReader.ReadInt16(), WorkingReader.ReadInt16()), new Vector3Short(WorkingReader.ReadInt16(), WorkingReader.ReadInt16(), WorkingReader.ReadInt16()));

				this.faceIndex = WorkingReader.ReadUInt16();
				this.faceCount = WorkingReader.ReadUInt16();


			}

			#endregion


			public override string ToString() {
				return (this.FrontIsNode ? "N" : "L") + this.FrontIndex + "/" + (this.BackIsNode ? "N" : "L") + this.BackIndex;
			}


		}


	}
}
