using System;
using System.IO;
using System.ComponentModel;

namespace id.Quake.Files {

	[SupportedExtension("pcx")]
	public class ZSoftPcx : IPicture, INamedResource {

		#region Properties

		private readonly int xMin;
		/// <summary>Gets the minimum X coordinate.</summary>
		[Description("Minimum X-coordinate.")]
		[Category("Bounds")]
		public int XMin {
			get { return this.xMin; }
		}

		private readonly int yMin;
		/// <summary>Gets the minimum y coordinate.</summary>
		[Description("Minimum Y-coordinate.")]
		[Category("Bounds")]
		public int YMin {
			get { return this.yMin; }
		}

		private readonly int xMax;
		/// <summary>Gets the maximum X coordinate.</summary>
		[Description("Maximum X-coordinate.")]
		[Category("Bounds")]
		public int XMax {
			get { return this.xMax; }
		}

		private readonly int yMax;
		/// <summary>Gets the maximum y coordinate.</summary>
		[Description("Maximum Y-coordinate.")]
		[Category("Bounds")]
		public int YMax {
			get { return this.yMax; }
		}

		private readonly byte bitsPerPixel;
		/// <summary>Gets the number of bits used to represent a pixel per plane.</summary>
		[Description("Number of bits used to represent a pixel per plane.")]
		[Category("Colour")]
		public byte BitsPerPixel {
			get { return this.bitsPerPixel; }
		}

		private readonly ushort horizontalResolution;
		/// <summary>Gets the horizontal resolution in dots per inch.</summary>
		[Description("Horizontal resolution in dots per inch.")]
		[Category("Resolution")]
		public ushort HorizontalResolution {
			get { return this.horizontalResolution; }
		}

		private readonly ushort verticalResolution;
		/// <summary>Gets the vertical resolution in dots per inch.</summary>
		[Description("Vertical resolution in dots per inch.")]
		[Category("Resolution")]
		public ushort VerticalResolution {
			get { return this.verticalResolution; }
		}


		private readonly byte planeCount;
		/// <summary>Gets the number of bits used to represent a pixel per plane.</summary>
		[Description("Number of colour planes in the image.")]
		[Category("Colour")]
		public byte PlaneCount {
			get { return this.planeCount; }
		}

		private readonly int[] palette;
		/// <summary>Get the palette used by the image as an array of 32-bpp ARGB integers.</summary>
		[Description("The palette used by the image as an array of 32-bpp ARGB integers.")]
		[Browsable(false)]
		public int[] Palette {
			get { return this.palette; }
		}

		private byte[] indexedPixels;
		/// <summary>Gets an array containing the palette indices of the image.</summary>
		[Browsable(false)]
		public byte[] IndexedPixels {
			get {
				if (this.indexedPixels == null) throw new InvalidOperationException("This is not a paletted image.");
				return this.indexedPixels; 
			}
		}

		private string name;
		[Browsable(false)]
		public string Name { get { return this.name; } set { this.name = value; } }

		[Browsable(false)]
		public bool NamedInternally { get { return false; } }

		#endregion

		#region Extended Properties

		public int Width { get { return (this.xMax - this.xMin) + 1; } }
		public int Height { get { return (this.yMax - this.yMin) + 1; } }

		#endregion

		#region Private Members

		private readonly ResourceLoader Resources;

		private readonly ushort BytesPerLine;

		private readonly Picture Data;

		#endregion

		#region Constructor

		public ZSoftPcx(ResourceLoader resources, Stream stream) {

			this.Resources = resources;

			BinaryReader PcxReader = new BinaryReader(stream);

			#region Header

			// Check manufacturer code:
			if (PcxReader.ReadByte() != 10) throw new InvalidDataException("Unsupported manufacturer code.");

			// Version?
			byte Version = PcxReader.ReadByte();

			// Get compression scheme:
			byte Compression;
			if ((Compression = PcxReader.ReadByte()) > 1) throw new InvalidDataException("Encoding not supported.");

			// Bits-per-pixel-per-plane:
			if ((this.bitsPerPixel = PcxReader.ReadByte()) != 8) throw new InvalidDataException("Only 8bpp images are supported.");

			// Skip window:
			this.xMin = PcxReader.ReadInt16();
			this.yMin = PcxReader.ReadInt16();
			this.xMax = PcxReader.ReadInt16();
			this.yMax = PcxReader.ReadInt16();

			// Resolution:
			this.horizontalResolution = PcxReader.ReadUInt16();
			this.verticalResolution = PcxReader.ReadUInt16();

			// Colour map:
			this.palette = new int[16];
			for (int i = 0; i < 16; ++i) this.palette[i] = unchecked((PcxReader.ReadByte() << 16) | (PcxReader.ReadByte() << 8) | (PcxReader.ReadByte() << 0) | (int)(0xFF000000));

			// Reserved:
			if (PcxReader.ReadByte() != 0) throw new InvalidDataException("Reserved field should be 0.");

			// Colour planes:
			if ((this.planeCount = PcxReader.ReadByte()) != 1) throw new InvalidDataException("Only single-plane images are supported.");

			// Number of bytes per scanline:
			this.BytesPerLine = PcxReader.ReadUInt16();

			#endregion

			#region Extended Palette

			if (Version == 5) {
				stream.Seek(-768, SeekOrigin.End);
				this.palette = new int[256];
				for (int i = 0; i < 256; ++i) Palette[i] = unchecked((PcxReader.ReadByte() << 16) | (PcxReader.ReadByte() << 8) | (PcxReader.ReadByte() << 0) | (int)(0xFF000000));
			}

			#endregion

			// Seek to the data itself:
			stream.Seek(128, SeekOrigin.Begin);

			#region Decompression

			int Width = this.XMax - this.XMin + 1;
			int Height = this.YMax - this.YMin + 1;
			byte[] Decompressed = new byte[Width * Height];

			// Storage for temporary scanline!
			//int[] Scanline = new int[this.PlaneCount * this.BytesPerLine];
			byte[] RawScanline = new byte[this.PlaneCount * this.BytesPerLine];

			int RunCount = 0;
			byte RunValue = 0;

			int DecompressedY = 0;
			int ScanlineX = 0;
			
			while (DecompressedY < Height) {

				// Read a single byte:
				byte Value = 0;
				if (--RunCount > 0) {
					Value = RunValue;
				} else {
					Value = PcxReader.ReadByte();
					if ((Value & 0xC0) == 0xC0) {
						RunCount = (byte)(Value & 0x3F);
						RunValue = PcxReader.ReadByte();
						Value = RunValue;
					}
				}

				RawScanline[ScanlineX++] = Value;
				if (ScanlineX == RawScanline.Length) {
					ScanlineX = 0;
					Array.Copy(RawScanline, 0, Decompressed, DecompressedY * Width, Width);
					++DecompressedY;
				}
			}
			
			#endregion
			this.indexedPixels = Decompressed;
			this.Data = new Picture(Width, Height, Array.ConvertAll<byte, int>(Decompressed, delegate(byte b) { return this.palette[b]; }));

		}

		#endregion

		#region IPicture Members

		public Picture Picture {
			get {
				return this.Data;
			}
		}

		#endregion
	}
}
