using System;
using System.Collections.Generic;
using System.Text;

namespace id.Quake.Files {
	public partial class QuakeC {

		#region Debug Strings

		public class DebugStringPrintedEventArgs : EventArgs {
			private string message;
			/// <summary>Gets the message.</summary>
			public string Message {
				get { return this.message; }
			}
			public DebugStringPrintedEventArgs(string message) {
				this.message = message;
			}

			public override string ToString() {
				return this.message;
			}
		}

		public delegate void DebugStringPrintedEventHandler(object sender, DebugStringPrintedEventArgs e);

		/// <summary>Event that is fired when a debug string is printed by the VM.</summary>
		public event DebugStringPrintedEventHandler DebugStringPrinted;
		protected virtual void OnDebugStringPrinted(DebugStringPrintedEventArgs e) {
			if (this.DebugStringPrinted != null) this.DebugStringPrinted(this, e);
		}

		#endregion

		#region Precache File

		public class PrecacheFileRequestedEventArgs : EventArgs {
			private int index;
			/// <summary>Gets the index used by Quake to reference the file.</summary>
			public int Index {
				get { return this.index; }
			}
			private string filename;
			/// <summary>Gets the filename.</summary>
			public string Filename {
				get { return this.filename; }
			}
			public PrecacheFileRequestedEventArgs(int index, string filename) {
				this.index = index;
				this.filename = filename;
			}

			public override string ToString() {
				return this.filename;
			}
		}

		public delegate void PrecacheFileRequestedEventHandler(object sender, PrecacheFileRequestedEventArgs e);

		/// <summary>Event that is fired when the application requests a file is precached.</summary>
		public event PrecacheFileRequestedEventHandler PrecacheFileRequested;
		protected virtual void OnPrecacheFileRequested(PrecacheFileRequestedEventArgs e) {
			if (this.PrecacheFileRequested != null) this.PrecacheFileRequested(this, e);
		}

		/// <summary>Event that is fired when the application requests a sound is precached.</summary>
		public event PrecacheFileRequestedEventHandler PrecacheSoundRequested;
		protected virtual void OnPrecacheSoundRequested(PrecacheFileRequestedEventArgs e) {
			if (this.PrecacheSoundRequested != null) this.PrecacheSoundRequested(this, e);
		}

		/// <summary>Event that is fired when the application requests a model is precached.</summary>
		public event PrecacheFileRequestedEventHandler PrecacheModelRequested;
		protected virtual void OnPrecacheModelRequested(PrecacheFileRequestedEventArgs e) {
			if (this.PrecacheModelRequested != null) this.PrecacheModelRequested(this, e);
		}

		#endregion

		#region CVar

		public class CVarChangedEventArgs : EventArgs {
			private string name;
			/// <summary>Gets the name of the variable.</summary>
			public string Name {
				get { return this.name; }
			}

			private string value;
			/// <summary>Gets the value of the variable.</summary>
			public string Value {
				get { return this.value; }
			}

			public CVarChangedEventArgs(string name, string value) {
				this.name = name; this.value = value;
			}

			public override string ToString() {
				return "{" + this.name + "=" + this.value + "}";
			}
		}

		public delegate void CVarChangedEventHandler(object sender, CVarChangedEventArgs e);
		/// <summary>Event that is fired when the application changes a cvar.</summary>
		public event CVarChangedEventHandler CVarChanged;
		protected virtual void OnCVarChanged(CVarChangedEventArgs e) {
			if (this.CVarChanged != null) this.CVarChanged(this, e);
		}

		#endregion

		#region Light Styles

		public class LightStyleChangedEventArgs : EventArgs {
			private int light;
			/// <summary>Gets the light's index.</summary>
			public int Light {
				get { return this.light; }
			}

			private string value;
			/// <summary>Gets the value of the light style.</summary>
			public string Value {
				get { return this.value; }
			}

			/// <summary>Gets the light levels as an array of bytes.</summary>
			public float[] Levels {
				get {
					return Array.ConvertAll<char, float>(this.value.ToCharArray(), delegate(char c) { return (float)(c - 'a') * (22f / 256f); });
				}
			}

			public LightStyleChangedEventArgs(int light, string value) {
				this.light = light; this.value = value;
			}

			public override string ToString() {
				return "{" + this.light + "=" + this.value + "}";
			}
		}

		public delegate void LightStyleChangedEventHandler(object sender, LightStyleChangedEventArgs e);
		/// <summary>Event that is fired when the application changes a cvar.</summary>
		public event LightStyleChangedEventHandler LightStyleChanged;
		protected virtual void OnLightStyleChanged(LightStyleChangedEventArgs e) {
			if (this.LightStyleChanged != null) this.LightStyleChanged(this, e);
		}

		#endregion

		#region Ambient Sound

		public class SpawnAmbientSoundEventArgs : EventArgs {

			private Vector3 position;
			public Vector3 Position {
				get { return this.position; }
			}

			private string name;
			public string Name {
				get { return this.name; }
			}

			private float volume;
			public float Volume {
				get { return this.volume; }
			}

			private float attenuation;
			public float Attenuation {
				get { return this.attenuation; }
			}


			public SpawnAmbientSoundEventArgs(Vector3 position, string name, float volume, float attenuation) {
				this.position = position; this.name = name; this.volume = volume; this.attenuation = attenuation;
			}

			public override string ToString() {
				return this.Name + "@" + this.Position + "*" + this.Volume + "/" + this.Attenuation;
			}
		}

		public delegate void SpawnAmbientSoundEventHandler(object sender, SpawnAmbientSoundEventArgs e);
		/// <summary>Event that is fired when the application spawns an ambient sound.</summary>
		public event SpawnAmbientSoundEventHandler SpawnAmbientSound;
		protected virtual void OnSpawnAmbientSound(SpawnAmbientSoundEventArgs e) {
			if (this.SpawnAmbientSound != null) this.SpawnAmbientSound(this, e);
		}

		#endregion

		#region Static Entities

		public class SpawnStaticEntityEventArgs : EventArgs {

			private EdictCollection.Edict entity;
			public EdictCollection.Edict Entity {
				get { return this.entity; }
			}

			public SpawnStaticEntityEventArgs(EdictCollection.Edict entity) {
				this.entity = entity;
			}

			public override string ToString() {
				return this.entity.ToString();
			}
		}

		public delegate void SpawnStaticEntityEventHandler(object sender, SpawnStaticEntityEventArgs e);
		/// <summary>Event that is fired when the application spawns a static (unreferenced) entity.</summary>
		public event SpawnStaticEntityEventHandler SpawnStaticEntity;
		protected virtual void OnSpawnStaticEntity(SpawnStaticEntityEventArgs e) {
			if (this.SpawnStaticEntity != null) this.SpawnStaticEntity(this, e);
		}

		#endregion

	}
}
