using System;
using QuakeC = id.Quake.Files.QuakeC;
using System.IO;
using System.Text;
using System.Collections.Generic;

namespace id.Quake {

	/// <summary>Defines a collection of edicts.</summary>
	public class EdictCollection {

		#region Private/Internal Members

		/// <summary>Reference to the QuakeC source.</summary>
		internal readonly Files.QuakeC QuakeCSource;

		/// <summary>Data storage for all edicts.</summary>
		internal Files.QuakeC.VariableValue[] EdictData;

		/// <summary>Array used to specify whether each edict is "alive" (and active).</summary>
		private bool[] Active;


		#endregion

		#region Constructor

		public EdictCollection(Files.QuakeC source) {
			this.QuakeCSource = source;
			this.Clear();
		}

		#endregion

		#region Allocation

		/// <summary>Allocate a new edict (or reuse an old erased one).</summary>
		public Edict Allocate() {
			// Check if there are any inactive edicts that can be re-used.
			for (int i = 0; i < this.Active.Length; ++i) {
				if (!this.Active[i]) {
					// This bugger is not used, so re-use it.
					this.Active[i] = true;
					return new Edict(this, i);
				}
			}

			// If we get this far, there aren't any free edicts.
			{
				Array.Resize<bool>(ref this.Active, this.Active.Length + 1);
				this.Active[this.Active.Length - 1] = true;
				Array.Resize<QuakeC.VariableValue>(ref this.EdictData, this.EdictData.Length + this.QuakeCSource.EdictSize);
				return new Edict(this, this.Active.Length - 1);
			}
		}

		/// <summary>Free an edict.</summary>
		public void Free(Edict edict) {
			this.Active[edict.Index] = false;
			// Erase any existing data:
			for (int j = 0, i = edict.Index * this.QuakeCSource.EdictSize; j < this.QuakeCSource.EdictSize; j++) {
				this.EdictData[i + j].Integer = 0;
			}
		}

		/// <summary>Clear all edicts.</summary>
		public void Clear() {
			this.EdictData = new QuakeC.VariableValue[] { };
			this.Active = new bool[] { };
		}

		/// <summary>Allocate a string and return its string pointer.</summary>
		/// <param name="value">The string to allocate.</param>
		private int AllocateString(string value) {
			int StringPointer = this.QuakeCSource.EdictStringData.Length;
			Array.Resize<byte>(ref this.QuakeCSource.EdictStringData, this.QuakeCSource.EdictStringData.Length + value.Length + 1);
			Array.Copy(Encoding.ASCII.GetBytes(value), 0, this.QuakeCSource.EdictStringData, StringPointer, value.Length);
			return unchecked(StringPointer | (int)0x80000000);
		}

		/// <summary>Load edicts from a BSP level.</summary>
		/// <param name="level">The level to load edicts from.</param>
		/// <param name="vm">An instance of a virtual machine to use to spawn edicts, or null if one is unavailable.</param>
		public void LoadFromLevel(Files.Level level, VirtualMachine vm) {

			// Erase existing items.
			this.Clear();

			this.QuakeCSource.EdictStringData = new byte[] { 0 };

			// Keep track of braces in source.
			int BraceDepth = 0;

			string EdictSourceText = level.Entities.Text.Split('\0')[0];

			Edict WorkingEntity = null;

			#region Parse Entity Data

			// Use these to accelerate accessing fields.
			QuakeC.Definition ClassName = this.QuakeCSource.Fields["classname"];
			QuakeC.Definition SpawnFlags = this.QuakeCSource.Fields["spawnflags"];
			
			// Edict inhibition via skill and mode:
			int SkillLevel = (int)(float)(this.QuakeCSource.Globals["skill"].Value);
			bool Deathmatch = (float)this.QuakeCSource.Globals["deathmatch"].Value != 0f;

			// Nicked from server.h
			const int SPAWNFLAG_NOT_EASY = 256;
			const int SPAWNFLAG_NOT_MEDIUM = 512;
			const int SPAWNFLAG_NOT_HARD = 1024;
			const int SPAWNFLAG_NOT_DEATHMATCH = 2048;


			// Iterate over all lines in source.
			foreach (string SourceLine in EdictSourceText.Split('\n')) {

				// Trim whitespace:
				string Line = SourceLine.Trim();

				// Skip empty lines:
				if (string.IsNullOrEmpty(Line)) continue;


				switch (Line) {
					case "{":
						++BraceDepth;
						if (BraceDepth > 2) throw new InvalidDataException("Too many opening braces.");
						switch (BraceDepth) {
							case 1: // New object!
								if (WorkingEntity != null) throw new InvalidDataException();
								WorkingEntity = Allocate();
								break;
						}
						break;
					case "}":
						switch (BraceDepth) {
							case 0: // Uh-oh.
								throw new InvalidDataException("Too many closing braces.");
							case 1: // Finished object.

								//HACK: Is this intended behaviour?
								if ((string)WorkingEntity.Properties[ClassName] == "worldspawn") {
									WorkingEntity.Properties["model"] = this.AllocateString(level.Name);
								}

								// Is it spawnable?
								int EdictSpawnFlags = (int)(float)WorkingEntity.Properties[SpawnFlags];

								// Almost identical code layout to pr_edict.c, sorry :|
								bool Inhibited = false;
								if (Deathmatch) {
									Inhibited = (EdictSpawnFlags & SPAWNFLAG_NOT_DEATHMATCH) != 0;
								} else {
									Inhibited =
										(SkillLevel == 0 && (EdictSpawnFlags & SPAWNFLAG_NOT_EASY) != 0) ||
										(SkillLevel == 1 && (EdictSpawnFlags & SPAWNFLAG_NOT_MEDIUM) != 0) ||
										(SkillLevel >= 2 && (EdictSpawnFlags & SPAWNFLAG_NOT_HARD) != 0);
								}

								if (Inhibited) {
									// This entity has been inhibited, so release it into the wild.
									this.Free(WorkingEntity);
								} else {
									// Call spawn function if available and there's an attached VM.
									if (vm != null) {
										string EntityClass = (string)WorkingEntity.Properties[ClassName];
										if (!string.IsNullOrEmpty(EntityClass)) {
											// Call the spawn function.
											vm.Invoke(WorkingEntity, EntityClass);
										}
									}
								}

								// Clear out the reference to the working entity.
								WorkingEntity = null;
								break;
						}
						--BraceDepth;

						if (BraceDepth < 0) throw new InvalidDataException("Too many closing braces.");
						break;
					default:
						switch (BraceDepth) {
							case 1: // Attributes.
								string[] Data = Line.Split('"');
								if (Data.Length != 5) throw new InvalidDataException("Couldn't parse property '" + Line + "'.");
								string Key = Data[1].TrimEnd(' '); // Trailing spaces hack.
								string Value = Data[3];

								//BOTCH: angle->angles.
								bool AngleHack = Key == "angle";
								if (AngleHack) {
									Key = "angles";
									Value = string.Format("0 {0} 0", Value);
								}

								//BOTCH: light->light_lev
								if (Key == "light") Key = "light_lev";

								if (Key[0] == '_') continue; //BOTCH: "Keynames with a leading underscore are used for utility comments, and are immediately discarded by Quake"

								// Get the type:
								QuakeC.Definition Def = this.QuakeCSource.Fields[Key];
								switch (Def.Type) {
									case QuakeC.Definition.DataType.String:
										WorkingEntity.Properties[Key] = this.AllocateString(Value);
										break;
									case QuakeC.Definition.DataType.Float:
										WorkingEntity.Properties[Key] = float.Parse(Value);
										break;
									case QuakeC.Definition.DataType.Vector:
										WorkingEntity.Properties[Key] = Vector3.Parse(Value);
										break;
									default:
										throw new InvalidDataException();
								}

								break;
							case 2: // Brush stuff.
								break;
						}
						break;
				}
			#endregion

			}
		}

		#endregion

		#region Indexers

		/// <summary>Retrieve an edict via its index.</summary>
		/// <param name="index">The index of the edict to retrieve.</param>
		public Edict this[int index] {
			get {
				if (!this.Active[index]) throw new ArgumentException("Edict " + index + " has been freed.");
				return new Edict(this, index);
			}
		}

		/// <summary>Retrieve an edict via its classname.</summary>
		/// <param name="classname">The classname of the edict to retrieve.</param>
		/// <remarks>An exception is thrown if there isn't exactly one edict with a matching classname.</remarks>
		public Edict this[string classname] {
			get {
				// Get ALL matching edicts.
				Edict[] Results = this.GetEdictByClassName(classname);

				// Only return an edict if one - and one only - was found.
				switch (Results.Length) {
					case 1:
						return Results[0];
					case 0:
						throw new InvalidOperationException("No edicts with the classname '" + classname + "' could be found.");
					default:
						throw new InvalidOperationException("More than one edict with the classname '" + classname + "' found.");
				}
			}
		}
		
		#endregion

		#region Public Methods

		/// <summary>Gets an array of edicts with the matching classname.</summary>
		/// <param name="classname">The name of the edict class to search for.</param>
		public Edict[] GetEdictByClassName(string classname) {

			// Storage for results.
			List<Edict> Results = new List<Edict>(this.Active.Length);

			// Look up the "classname" field definition.
			QuakeC.Definition ClassNameField = this.QuakeCSource.Fields["classname"];

			// Iterate over all active edicts:
			for (int i = 0, j = 0; i < this.Active.Length; ++i, j += this.QuakeCSource.EdictSize) {
				if (this.Active[i]) { // Is it active?

					// Check the classname.
					if (this.QuakeCSource.GetString(this.EdictData[j + ClassNameField.Offset].Integer) == classname) {
						Results.Add(new Edict(this, i));
					}
					
				}
			}

			// Return results.
			return Results.ToArray();

		}

		/// <summary>Gets an array of all active edicts.</summary>
		public Edict[] GetActiveEdicts() {
			// Storage for results.
			List<Edict> Results = new List<Edict>(this.Active.Length);

			// Iterate over all active edicts:
			for (int i = 0, j = 0; i < this.Active.Length; ++i, j += this.QuakeCSource.EdictSize) {
				if (this.Active[i]) { // Is it active?
					Results.Add(new Edict(this, i));
				}
			}

			// Return results.
			return Results.ToArray();
		}

		#endregion

		#region Properties

		/// <summary>Gets the maximum number of edicts that have been allocated in this collection.</summary>
		/// <remarks>
		/// When an edict is freed its memory can be reclaimed by a subsequent allocation;
		/// if no freed edicts are available for overwriting then the capacity of the collection is enlarged.
		/// </remarks>
		public int Capacity {
			get { return this.Active.Length; }
		}

		/// <summary>Check whether an edict's index is active.</summary>
		/// <param name="index">The index of the edict to check.</param>
		/// <returns>True if the edict is active, false if the space has been marked for reuse by a subsequent allocation.</returns>
		public bool IsActive(int index) {
			return this.Active[index];
		}

		#endregion

	}
}
