using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;

namespace id.Quake {

	/// <summary>Defines a plane.</summary>
	public struct Plane {

		#region Types

		/// <summary>Describes the side of a plane something is.</summary>
		public enum Side {
			/// <summary>The item is in front of the plane.</summary>
			Front,
			/// <summary>The item is in behind the plane.</summary>
			Back,
			/// <summary>The item lies inside the plane.</summary>
			Inside,
		}

		#endregion

		#region Fields

		/// <summary>Gets or sets the plane normal.</summary>
		public Vector3 Normal;

		/// <summary>Gets or sets the offset of the plane from the origin.</summary>
		public float Offset;

		#endregion

		#region Constructor

		/// <summary>Creates an instance of the plane from a normal and offset value.</summary>
		/// <param name="normal">The plane's normal.</param>
		/// <param name="offset">The offset of the plane along the normal from the origin.</param>
		public Plane(Vector3 normal, float offset) {
			this.Normal = normal;
			this.Offset = offset;
		}

		#endregion

		#region Conversions

		/// <summary>Explicit conversion operator used to cast from a BSP level plane.</summary>
		public static explicit operator Plane(Quake.Files.Level.Plane p) {
			return new Plane(p.Normal, p.Offset);
		}
		
		#endregion

		#region Methods

		/// <summary>Get the side of the plane a point lies in.</summary>
		/// <param name="p">The position of the point to check.</param>
		public Side GetSide(Vector3 p) {
			float f = Vector3.Dot(this.Normal, p) - this.Offset;
			if (f < 0) {
				return Side.Front;
			} else if (f > 0) {
				return Side.Back;
			} else {
				return Side.Inside;
			}
		}

		/// <summary>Calculate the collision of a point moving through the plane.</summary>
		/// <param name="start">The starting position of the point.</param>
		/// <param name="end">The desired end position of the point.</param>
		public CollisionResult<Vector3> GetCollision(Vector3 start, Vector3 end) {

			// Get the distance along of the intersection:
			Vector3 FacingNormal = -this.Normal;
			float Distance = -(Vector3.Dot(start, FacingNormal) + this.Offset) / Vector3.Dot(end - start, FacingNormal);

			// Check range:
			if (Distance < 0 || Distance > 1) {
				// Intersection point happens behind or in front of the ends of the line.
				// This means that either the line isn't long enough to make contact with the plane, or that the 
				// line doesn't hit the plane at all (ie, it lies parallel to it).
				return new CollisionResult<Vector3>(false, end);
			} else {
				// We intersect the line somewhere along its length.
				return new CollisionResult<Vector3>(true, start + (end - start) * Distance, Distance, this);
			}
		}

		/// <summary>Checks whether a bounding box intersects the plane.</summary>
		/// <param name="box">The bounding box to check.</param>
		/// <returns>True if the bounding box intersects the plane, false otherwise.</returns>
		public bool Intersects(BoundingBox box) {
			Side LastSide = default(Side);
			Vector3[] Corners = box.Corners;
			for (int i = 0; i < Corners.Length; ++i) {
				Side ThisSide = this.GetSide(Corners[i]);
				if (i == 0) LastSide = ThisSide;
				if (LastSide != ThisSide) return true;
				LastSide = ThisSide;
			}
			return false;
		}

		/// <summary>Get the collision position between a bounding box and a plane.</summary>
		/// <param name="box">The bounding box to move.</param>
		/// <param name="direction">The direction vector to move the box by.</param>
		public CollisionResult<BoundingBox> GetCollision(BoundingBox box, Vector3 direction) {

			// Find the end position:
			BoundingBox End = box + direction;

			// Quick bounding-box intersection test:
			if (!this.Intersects(BoundingBox.Combine(box, End))) {
				return new CollisionResult<BoundingBox>(false, End);
			}

			// So, we hit the plane somewhere:
			Vector3[] Starts = box.Corners;
			Vector3[] Ends = End.Corners;

			//OPTIMISATION: Only check corners on sides that are moving in direction of bounding box.

			// Perform collision detection for all eight corners:
			CollisionResult<BoundingBox> Result = new CollisionResult<BoundingBox>(false, End);
			for (int i = 0; i < 8; ++i) {
				CollisionResult<Vector3> C = this.GetCollision(Starts[i], Ends[i]);
				if (C.CollisionOccured && C.Distance < Result.Distance) {
					Result.CollisionOccured = true;
					Result.Distance = C.Distance;
				}
			}

			// If a collision occured...
			if (Result.CollisionOccured) {
				Result.Plane = this;
				Result.Position = box + direction * Result.Distance;
			}

			return Result;

		}

		#endregion





	}
}
