using System;
using QuakeC = id.Quake.Files.QuakeC;
using Statement = id.Quake.Files.QuakeC.Statement;

namespace id.Quake {
	public partial class VirtualMachine {

		/// <summary>Execute the body of a function.</summary>
		/// <param name="function">The function to execute.</param>
		/// <remarks>Only returns once the inner </remarks>
		private void ExecuteFunctionBody(QuakeC.Function function) {

			// Remember the stack level before we entered the function.
			int EntryStackLevel = this.FunctionStack.Count;

			// Enter a function:
			this.EnterFunction(function);

			// Infinite loop (we'll exit when the QuakeC code tells us to).
			for (; ; ) {
				// Fetch the statement to execute.
				QuakeC.Statement Statement = this.quakeCSource.Statements[StatementPointer++];

				WorkingVariable A = new WorkingVariable(this, Statement.OperandIndices[0]);
				WorkingVariable B = new WorkingVariable(this, Statement.OperandIndices[1]);
				WorkingVariable C = new WorkingVariable(this, Statement.OperandIndices[2]);

				switch (Statement.OperationCode) {
					case Statement.Operation.OP_ADD_F:
						C.Float = A.Float + B.Float;
						break;
					case Statement.Operation.OP_ADD_V:
						C.Vector = A.Vector + B.Vector;
						break;
					case Statement.Operation.OP_SUB_F:
						C.Float = A.Float - B.Float;
						break;
					case Statement.Operation.OP_SUB_V:
						C.Vector = A.Vector - B.Vector;
						break;
					case Statement.Operation.OP_MUL_F:
						C.Float = A.Float * B.Float;
						break;
					case Statement.Operation.OP_MUL_V: {
							Vector3 a = A.Vector; Vector3 b = B.Vector;
							C.Float = a.X * b.X + a.Y * b.Y + a.Z * b.Z;
						}
						break;
					case Statement.Operation.OP_MUL_FV:
						C.Vector = A.Float * B.Vector;
						break;
					case Statement.Operation.OP_MUL_VF:
						C.Vector = B.Float * A.Vector;
						break;
					case Statement.Operation.OP_DIV_F:
						C.Float = A.Float / B.Float;
						break;
					case Statement.Operation.OP_BITAND:
						C.Float = (int)A.Float & (int)B.Float;
						break;
					case Statement.Operation.OP_BITOR:
						C.Float = (int)A.Float | (int)B.Float;
						break;
					case Statement.Operation.OP_GE:
						C.Float = (A.Float >= B.Float) ? 1f : 0f;
						break;
					case Statement.Operation.OP_LE:
						C.Float = (A.Float <= B.Float) ? 1f : 0f;
						break;
					case Statement.Operation.OP_GT:
						C.Float = (A.Float > B.Float) ? 1f : 0f;
						break;
					case Statement.Operation.OP_LT:
						C.Float = (A.Float < B.Float) ? 1f : 0f;
						break;
					case Statement.Operation.OP_AND:
						C.Boolean = A.Boolean && B.Boolean;
						break;
					case Statement.Operation.OP_OR:
						C.Boolean = A.Boolean || B.Boolean;
						break;
					case Statement.Operation.OP_NOT_F:
						C.Boolean = !A.Boolean;
						break;
					case Statement.Operation.OP_NOT_V:
						C.Boolean = (A.Vector == Vector3.Zero);
						break;
					case Statement.Operation.OP_NOT_S:
						C.Boolean = A.Integer != 0 || !string.IsNullOrEmpty(A.String);
						break;
					case Statement.Operation.OP_NOT_FNC:
						C.Boolean = A.Integer != 0;
						break;
					case Statement.Operation.OP_NOT_ENT:
						C.Boolean = A.Integer != 0; // 0 = world entity.
						break;
					case Statement.Operation.OP_EQ_F:
						C.Boolean = A.Float == B.Float;
						break;
					case Statement.Operation.OP_EQ_V:
						C.Boolean = A.Vector == B.Vector;
						break;
					case Statement.Operation.OP_EQ_S:
						C.Boolean = A.String == B.String;
						break;
					case Statement.Operation.OP_EQ_E:
					case Statement.Operation.OP_EQ_FNC:
						C.Boolean = A.Integer == B.Integer;
						break;

					case Statement.Operation.OP_NE_F:
						C.Boolean = A.Float != B.Float;
						break;
					case Statement.Operation.OP_NE_V:
						C.Boolean = A.Vector != B.Vector;
						break;
					case Statement.Operation.OP_NE_S:
						C.Float = string.Compare(A.String, B.String);
						break;
					case Statement.Operation.OP_NE_E:
					case Statement.Operation.OP_NE_FNC:
						C.Boolean = A.Integer != B.Integer;
						break;

					case Statement.Operation.OP_STORE_F:
					case Statement.Operation.OP_STORE_S:
					case Statement.Operation.OP_STORE_ENT:
					case Statement.Operation.OP_STORE_FLD:
					case Statement.Operation.OP_STORE_FNC:
						B.Integer = A.Integer;
						break;
					case Statement.Operation.OP_STORE_V:
						B.Vector = A.Vector;
						break;

					case Statement.Operation.OP_IFNOT:
						if (!A.Boolean) StatementPointer += (short)Statement.OperandIndices[1] - 1;
						break;
					case Statement.Operation.OP_IF:
						if (A.Boolean) StatementPointer += (short)Statement.OperandIndices[1] - 1;
						break;
					case Statement.Operation.OP_GOTO:
						StatementPointer += (short)Statement.OperandIndices[0] - 1;
						break;


					case Statement.Operation.OP_STOREP_F:
					case Statement.Operation.OP_STOREP_ENT:
					case Statement.Operation.OP_STOREP_FLD:
					case Statement.Operation.OP_STOREP_S:
					case Statement.Operation.OP_STOREP_FNC:
						this.Edicts.EdictData[B.Integer].Integer = A.Integer;
						break;
					case Statement.Operation.OP_STOREP_V: {
							Vector3 V = A.Vector;
							this.Edicts.EdictData[B.Integer + 0].Float = V.X;
							this.Edicts.EdictData[B.Integer + 1].Float = V.Y;
							this.Edicts.EdictData[B.Integer + 2].Float = V.Z;
						}
						break;

					case Statement.Operation.OP_ADDRESS:
						C.Integer = (A.Edict.Index * this.quakeCSource.EdictSize) + B.Integer;
						break;

					case Statement.Operation.OP_LOAD_F:
					case Statement.Operation.OP_LOAD_FLD:
					case Statement.Operation.OP_LOAD_ENT:
					case Statement.Operation.OP_LOAD_S:
					case Statement.Operation.OP_LOAD_FNC:
						C.Integer = this.Edicts.EdictData[A.Edict.Index * this.quakeCSource.EdictSize + B.Integer].Integer;
						break;
					case Statement.Operation.OP_LOAD_V:
						C.Vector = new Vector3(
							this.Edicts.EdictData[A.Edict.Index * this.quakeCSource.EdictSize + B.Integer + 0].Float,
							this.Edicts.EdictData[A.Edict.Index * this.quakeCSource.EdictSize + B.Integer + 1].Float,
							this.Edicts.EdictData[A.Edict.Index * this.quakeCSource.EdictSize + B.Integer + 2].Float
						);
						break;

					case Statement.Operation.OP_CALL0:
					case Statement.Operation.OP_CALL1:
					case Statement.Operation.OP_CALL2:
					case Statement.Operation.OP_CALL3:
					case Statement.Operation.OP_CALL4:
					case Statement.Operation.OP_CALL5:
					case Statement.Operation.OP_CALL6:
					case Statement.Operation.OP_CALL7:
					case Statement.Operation.OP_CALL8:
						QuakeC.Function Func = A.Function;
						if (Func.FirstStatement < 0) {
							this.BuiltInFunctions[-Func.FirstStatement]();
						} else {
							this.ExecuteFunctionBody(Func);
						}
						break;

					case Statement.Operation.OP_DONE:
					case Statement.Operation.OP_RETURN: {
							Vector3 V = A.Vector;
							this.quakeCSource.GlobalData[(int)WorkingVariableOffsets.Return + 0].Float = A.Vector.X;
							this.quakeCSource.GlobalData[(int)WorkingVariableOffsets.Return + 1].Float = A.Vector.Y;
							this.quakeCSource.GlobalData[(int)WorkingVariableOffsets.Return + 2].Float = A.Vector.Z;
						}
						this.LeaveFunction();
						if (this.FunctionStack.Count == EntryStackLevel) return;
						break;

					case Statement.Operation.OP_STATE:
						//TODO: Check #ifdef FPS_20
						Edict SelfEdict = this.edicts[(int)this.Self.Value];
						SelfEdict.Properties["nextthink"] = (float)this.quakeCSource.Globals["time"].Value + 0.1f;
						SelfEdict.Properties["frame"] = A.Float;
						SelfEdict.Properties["think"] = B.Function; // Assign function.
						break;

					default:
						throw new InvalidOperationException("Operation " + Statement.OperationCode.ToString() + " not supported.");
				}


			}

		}

	}
}
